# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Definition of a masked array.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet
"""
import numpy as np


def _merge_masks(ma1, ma2):
    """Merge the masks of two :class:`MadArray` objects and return the
    arguments used for initialisation of the resulting :class:`MadArray`
    object.

    Parameters
    ----------
    ma1 : MadArray
       First masked array to consider.
    ma2 : MadArray
       Second masked array to consider.

    Returns
    -------
    dict
        Arguments to be used for the initialisation of a :class:`MadArray`
        object.
    """
    if ma1._complex_masking or ma2._complex_masking:

        if ma1._complex_masking:
            mm1 = ma1.get_unknown_mask('magnitude')
            mp1 = ma1.get_unknown_mask('phase')
        else:
            mm1 = ma1.get_unknown_mask('any')
            mp1 = ma1.get_unknown_mask('any')

        if ma2._complex_masking:
            mm2 = ma2.get_unknown_mask('magnitude')
            mp2 = ma2.get_unknown_mask('phase')
        else:
            mm2 = ma2.get_unknown_mask('any')
            mp2 = ma2.get_unknown_mask('any')

        mask_magnitude = np.logical_or(mm1, mm2)
        mask_phase = np.logical_or(mp1, mp2)

        return {'mask_magnitude': mask_magnitude, 'mask_phase': mask_phase}
    else:
        return {'mask': (ma1.get_unknown_mask()) | (ma2.get_unknown_mask())}


UFUNC_NOT_RETURNING_MADARRAYS = ['bitwise_and', 'bitwise_or', 'bitwise_xor',
                                 'invert', 'left_shift', 'right_shift',
                                 'greater', 'greater_equal', 'less',
                                 'less_equal', 'not_equal', 'equal',
                                 'logical_and', 'logical_or', 'logical_xor',
                                 'logical_not', 'maximum', 'minimum', 'fmax',
                                 'fmin', 'isfinite', 'isinf', 'isnan', 'isnat',
                                 'signbit', 'copysign', 'nextafter', 'spacing',
                                 'modf',  'frexp', 'fmod']


class MadArray(np.ndarray):
    """Subclass of numpy.ndarray to handle data with missing elements.

    .. _type_entry_madarray:

    **Type of entry**: entries of array can be *int*, *float*, or *complex*.

    .. _masking_madarray:

    **Masking**: the masking of entries has two different modes:

    * Entries can be either masked or not masked, leading to a boolean mask,
      whose entries are equal to True if the corresponding data entry is
      masked, or False otherwise.
      
      This is the default mode and the mode selected when specifying ``mask``
      during creation.

    * Complex entries can have only the magnitude or phase component masked, or
      both. The resulting mask has integers entries, equal to:

        * *0* if the phase and the magnitude are not masked (known magnitude
          and phase);
        * *1* if only the phase is masked (known magnitude, unknown phase);
        * *2* if only the magnitude is masked (unknown magnitude, known phase);
        * *3* if the magnitude and the phase are masked (unknown magnitude and
          phase).

      This mode is selected when specifying ``mask_magnitude`` and/or
      ``mask_phase`` during creation.
      Entries are converted to a complex type.

      If entries are complex values and ``mask`` is given during creation, both
      the magnitude and phase are masked and the boolean mask mode is used.

    .. _indexing_madarray:

    **Indexing**: two different modes to index a :class:`MadArray` object are
    implemented:

    * a :class:`MadArray` object with shape corresponding to the indices is
      returned, with both the data matrix and the mask properly indexed. This
      is the default mode;
    * a :class:`MadArray` object with unchanged shape is returned, where
      non-indexed entries are set as masked. This mode is selected by setting
      the parameter ``masked_indexing`` to True.

    .. _numpy_behaviour_madarray:

    **Numpy behaviour**: it is possible to use standard operations (+, -, /,
    //, \*, T) between two :class:`MadArray` objects, likewise operations
    between numpy arrays. The resulting object has a mask consisting of the
    union of the operands. It is also possible to use pickle operations to
    jointly store the data and the mask.

    Parameters
    ----------
    data : array_like
        Multidimensional array. See :ref:`Type of Entry<type_entry_madarray>`.
    mask : boolean array_like, optional
        Mask for boolean masking mode.
        See :ref:`Masking<masking_madarray>`.
    mask_magnitude : boolean array_like or None, optional
        Magnitude mask for masking with complex data.
        See :ref:`Masking<masking_madarray>`.
    mask_phase : boolean or array_like or None, optional
        Phase mask for masking with complex data.
        See :ref:`Masking<masking_madarray>`.
    masked_indexing : bool or None, optional
        Indicate how the indexing is performed. If None, set to False.
        See :ref:`Indexing<indexing_madarray>`.

    Warnings
    --------
    This class inherits from ndarray or subclass of ndarray. Instances can be
    then manipulated like ndarrays (e.g., indexation). While some methods have
    been implemented taking into account the mask, some may cause unexpected
    behavior (e.g., mean).

    See also
    --------
    :mod:`numpy.doc.subclassing`.

    Notes
    -----
    This class implements an alternative masked array different from
    :class:`numpy.ma.MaskedArray`. The reason of this choice is that it is only
    used as a container of a ndarray and a mask. No masked operations are
    needed.
    """
    def __new__(cls, data, mask=None, mask_magnitude=None, mask_phase=None,
                masked_indexing=None, **kwargs):

        if mask is not None and mask_magnitude is not None:
            errmsg = ('Parameters mask and mask_magnitude are mutually '
                'exclusive')
            raise ValueError(errmsg)

        if mask is not None and mask_phase is not None:
            errmsg = 'Parameters mask and mask_phase are mutually exclusive'
            raise ValueError(errmsg)

        _data = np.array(data, **kwargs)

        if not (np.issubdtype(_data.dtype, np.floating) or
                np.issubdtype(_data.dtype, np.integer) or
                np.issubdtype(_data.dtype, np.complexfloating)):
            errmsg = 'Invalid dtype: {}'
            raise TypeError(errmsg.format(data.dtype))

        if mask is not None:
            complex_masking = False
        elif mask_magnitude is not None or mask_phase is not None:
            complex_masking = True
        elif isinstance(data, MadArray):
            complex_masking = data._complex_masking
        else:
            complex_masking = False

        if masked_indexing is None:
            if isinstance(data, MadArray):
                masked_indexing = data._masked_indexing
            else:
                masked_indexing = False

        if not complex_masking:

            if mask is None:
                if isinstance(data, MadArray):
                    mask = data.get_unknown_mask()
                else:
                    mask = np.zeros(_data.shape, dtype=np.bool)
            else:
                mask = np.array(mask, dtype=np.bool)
                if mask.shape != _data.shape:
                    errmsg = "Mask shape {} and data shape {} not compatible."
                    raise ValueError(errmsg.format(mask.shape, data.shape))
                mask = mask

        else:

            if not np.issubdtype(_data.dtype, np.complexfloating):
                _data = _data.astype(np.complex)

            if mask_magnitude is None:
                if isinstance(data, MadArray) and mask_phase is None:
                    mask_magnitude = data.get_unknown_mask('magnitude')
                    mask_phase = data.get_unknown_mask('phase')
                else:
                    mask_magnitude = np.zeros_like(data, dtype=np.bool)
            else:
                mask_magnitude = np.array(mask_magnitude, dtype=np.bool)
                if mask_magnitude.shape != _data.shape:
                    errmsg = 'Magnitude mask shape {} and data shape {} not '\
                        'compatible.'
                    raise ValueError(errmsg.format(mask_magnitude.shape,
                                                   _data.shape))

            if mask_phase is None:
                mask_phase = np.zeros_like(data, dtype=np.bool)

            else:
                mask_phase = np.array(mask_phase, dtype=np.bool)
                if mask_phase.shape != _data.shape:
                    errmsg = 'Phase mask shape {} and data shape {} not '\
                        'compatible.'
                    raise ValueError(errmsg.format(
                        mask_phase.shape, _data.shape))

            mask = np.zeros(_data.shape, dtype=np.uint8)
            mask[np.logical_and(mask_phase, ~mask_magnitude)] = 1
            mask[np.logical_and(~mask_phase, mask_magnitude)] = 2
            mask[np.logical_and(mask_phase, mask_magnitude)] = 3

        # create the object
        obj = np.ndarray.__new__(cls, _data.shape, dtype=_data.dtype)
        obj[:] = _data
        obj._mask = mask
        obj._masked_indexing = masked_indexing
        obj._complex_masking = complex_masking

        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._mask = getattr(obj, '_mask', None)
        self._complex_masking = getattr(obj, '_complex_masking', None)
        self._masked_indexing = getattr(obj, '_masked_indexing', None)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):

        args = []
        is_mad = []
        for input_ in inputs:
            if isinstance(input_, MadArray):
                args.append(input_.view(np.ndarray))
                is_mad.append(True)
            else:
                args.append(input_)
                is_mad.append(False)

        if len(is_mad) > 1:
            if np.all(is_mad):
                if inputs[0]._complex_masking or inputs[1]._complex_masking:
                    errmsg = 'Operation not permitted when complex masking.'
                    raise ValueError(errmsg)

                mask = MadArray(inputs[0].data, **
                                _merge_masks(inputs[0], inputs[1]))._mask
                complex_masking = inputs[0]._complex_masking
                masked_indexing = inputs[0]._masked_indexing
            else:
                where_is_mad = np.argwhere(is_mad).squeeze()
                mask = inputs[where_is_mad]._mask
                complex_masking = inputs[where_is_mad]._complex_masking
                masked_indexing = inputs[where_is_mad]._masked_indexing

        else:
            mask = inputs[0]._mask
            complex_masking = inputs[0]._complex_masking
            masked_indexing = inputs[0]._masked_indexing

        outputs = kwargs.pop('out', None)
        if outputs:
            out_args = []
            for output in outputs:
                out_args.append(output.view(np.ndarray))
            kwargs['out'] = tuple(out_args)
        else:
            outputs = (None,) * ufunc.nout

        results = (super().__array_ufunc__(ufunc, method, *args, **kwargs), )

        new_results = []
        for result, output in zip(results, outputs):
            if output is None:
                if (method == '__call__' and
                        ufunc.__name__ not in UFUNC_NOT_RETURNING_MADARRAYS):
                    new_results.append(np.asarray(result).view(MadArray))
                    new_results[-1]._mask = mask
                    new_results[-1]._complex_masking = complex_masking
                    new_results[-1]._masked_indexing = masked_indexing
                else:
                    new_results.append(np.asarray(result).view(np.ndarray))
            else:
                new_results.append(output)
        results = tuple(new_results)

        return results[0] if len(results) == 1 else results

    def __getitem__(self, index):

        if (getattr(self, '_masked_indexing', None) is not None and
                self._masked_indexing):

            mask = np.zeros(self.shape, dtype=np.bool)
            mask[index] = True
            return MadArray(self, mask=np.logical_or(~mask, self._mask))
        else:
            out_arr = super().__getitem__(index)
            if getattr(out_arr, '_mask', None) is not None:
                out_arr._mask = out_arr._mask[index]
            return out_arr

    def __reduce__(self):
        pickled_state = super().__reduce__()
        new_state = pickled_state[2] + (self._mask, self._complex_masking,
                                        self._masked_indexing)
        return pickled_state[0], pickled_state[1], new_state

    def __setstate__(self, state):
        self._mask = state[-3]
        self._complex_masking = state[-2]
        self._masked_indexing = state[-1]
        super().__setstate__(state[0:-3])

    @property
    def n_missing_data(self):
        """Number of missing data (double or tuple).

        Number of masked coefficients if dtype is int or float. Number of
        masked coefficients in phase and magnitude masks if dtype is complex.
        """
        if self._complex_masking:
            return (np.sum(self.get_unknown_mask('magnitude')),
                    np.sum(self.get_unknown_mask('phase')))
        else:
            return np.sum(self.get_unknown_mask())

    @property
    def ratio_missing_data(self):
        """Ratio of missing data (double or tuple).

        Ratio of masked coefficients if dtype is int or float. Ratio of
        masked coefficients in phase and magnitude masks if dtype is complex.
        """
        if self._complex_masking:
            return (np.average(self.get_unknown_mask('magnitude')),
                    np.average(self.get_unknown_mask('phase')))
        else:
            return np.average(self.get_unknown_mask())

    def is_masked(self):
        """Indicate if one or several elements are masked."""
        return np.any(self._mask)

    def to_np_array(self, fill_value=None):
        """Return a numpy array.

        If ``fill_value`` is not None, masked elements are replaced according
        to the type of entries:

        * ``fill_value`` if the type of entries is *int* or *float*;
        * If the type is *complex*, missing entries are replaced either by:
            * a complex number with the known magnitude value without the phase
              information if only the phase is masked;
            * a complex number of magnitude 1 with the known phase if only the
              magnitude is masked;
            * by ``fill_value`` if both magnitude and phase are masked.

        Parameters
        ----------
        fill_value : scalar or None
            Value used to fill masked elements. If None, the initial value is
            kept.

        Returns
        -------
        nd-array
        """
        data = np.array(self)
        if fill_value is not None:
            if self._complex_masking:
                upom = self.get_unknown_mask('phase only')
                umom = self.get_unknown_mask('magnitude only')
                data[upom] = np.abs(data[upom])
                data[umom] = np.exp(1j * np.angle(data[umom]))
                data[self.get_unknown_mask('all')] = fill_value
            else:
                data[self.get_unknown_mask()] = fill_value
        return data

    def __eq__(self, other):
        if isinstance(other, MadArray):
            return np.logical_and(self.to_np_array(0) == other.to_np_array(0),
                                  self._mask == self._mask)
        else:
            return np.array(self) == other

    def __ne__(self, other):
        return np.logical_not(self == other)

    def is_equal(self, other):
        if not isinstance(other, MadArray):
            return False

        if not np.all(self == other):
            return False

        if not (self._complex_masking == other._complex_masking and
                self._masked_indexing == other._masked_indexing):
            return False

        return True

    @property
    def T(self):
        """Transpose of the MadArray."""
        return self.transpose()

    def copy(self):
        return MadArray(self)

    def transpose(self):
        mat = super().transpose()
        mat._mask = mat._mask.transpose()
        return mat

    def __str__(self):
        arr = np.array(self)

        if np.issubdtype(self.dtype, np.integer):
            arr = arr.astype(np.float64)

        arr[self.get_unknown_mask()] = np.nan
        arr_str = np.ndarray.__str__(arr)
        if np.isrealobj(arr):
            arr_str = arr_str.replace('nan', '  x')
        else:
            arr_str = arr_str.replace('nan+0.j', '  x    ')

        if np.issubdtype(self.dtype, np.integer):
            arr_str = arr_str.replace('.', '')

        if self._complex_masking:
            n_all_unknown = np.count_nonzero(self.get_unknown_mask('all'))
            string = 'MadArray, dtype={0}, ' \
                     '{1[0]} missing magnitudes ({2[0]:.1%}) ' \
                     'and {1[1]} missing phases ({2[1]:.1%}), ' \
                     ' including {3} missing magnitudes and phases jointly ' \
                     '({4:.1%})\n{5}'
            return string.format(self.dtype,
                                 self.n_missing_data,
                                 self.ratio_missing_data,
                                 n_all_unknown,
                                 n_all_unknown/self.size,
                                 arr_str)
        else:
            string = 'MadArray, dtype={}, {} missing entries ({:.1%})\n{}'

            return string.format(self.dtype,
                                 self.n_missing_data,
                                 self.ratio_missing_data, arr_str)

    def __repr__(self):
        string = '<MadArray at {}>'
        return string.format(hex(id(self)))

    def get_known_mask(self, mask_type='all'):
        """Boolean mask for known coefficients.

        Compute the boolean mask marking known coefficients as True.

        Parameters
        ----------
        mask_type : {'all', 'any', 'magnitude', 'phase', 'magnitude only', \
                    'phase only'}
            Type of mask:

            - ``all``: mark coefficients for wich both the magnitude and the
              phase are known,
            - ``any``: mark coefficients for wich the magnitude or the phase
              are known (including when both the magnitude and the phase are
              known),
            - ``magnitude``: mark coefficients for wich the magnitude is
              known,
            - ``phase``: mark coefficients for wich the phase is known,
            - ``magnitude only``: mark coefficients for wich both the magnitude
              is known and the phase is unknown,
            - ``phase only``: mark coefficients for wich both the phase is
              known and the magnitude is unknown.
            
        Returns
        -------
        mask : boolean nd-array
            Boolean array with entries set to True if the corresponding value
            in the object is known.

        Raises
        ------
        ValueError
            If ``mask_type`` has an invalid value.
        """
        if mask_type == 'all':
            return ~self.get_unknown_mask('any')
        elif mask_type == 'any':
            return ~self.get_unknown_mask('all')
        elif mask_type == 'magnitude':
            return ~self.get_unknown_mask('magnitude')
        elif mask_type == 'phase':
            return ~self.get_unknown_mask('phase')
        elif mask_type == 'magnitude only':
            return self.get_unknown_mask('phase only')
        elif mask_type == 'phase only':
            return self.get_unknown_mask('magnitude only')

        errmsg = 'Invalid value for mask_type: {}'.format(mask_type)
        raise ValueError(errmsg)

    def get_unknown_mask(self, mask_type='any'):
        """Boolean mask for unknown coefficients.

        Compute the boolean mask marking unknown coefficients as True.

        Parameters
        ----------
        mask_type : {'any', 'all', 'magnitude', 'phase', 'magnitude only', \
                    'phase only'}
            Type of mask:

            - ``any``: mark coefficients for wich the magnitude or the phase
              are unknown (including when both the magnitude and the phase are
              unknown),
            - ``all``: mark coefficients for wich both the magnitude and the
              phase are unknown,
            - ``magnitude``: mark coefficients for wich the magnitude is
              unknown,
            - ``phase``: mark coefficients for wich the phase is unknown,
            - ``magnitude only``: mark coefficients for wich both the magnitude
              is unknown and the phase is known,
            - ``phase only``: mark coefficients for wich both the phase is
              unknown and the magnitude is known.

        Returns
        -------
        mask : boolean nd-array
            Boolean array with values set to True if the corresponding value in
            the object is unknown.

        Raises
        ------
        ValueError
            If ``mask_type`` has an invalid value.
        """
        if self._complex_masking:
            if mask_type == 'any':
                return self._mask != 0
            elif mask_type == 'all':
                return self._mask == 3
            elif mask_type == 'magnitude':
                return (self._mask == 3) | (self._mask == 2)
            elif mask_type == 'phase':
                return (self._mask == 3) | (self._mask == 1)
            elif mask_type == 'magnitude only':
                return self._mask == 2
            elif mask_type == 'phase only':
                return self._mask == 1
        else:
            if mask_type in ('any', 'all', 'magnitude', 'phase'):
                return np.copy(self._mask)
            elif mask_type in ('magnitude only', 'phase only'):
                return np.zeros_like(self._mask, dtype=np.bool)

        errmsg = 'Invalid value for mask_type: {}'.format(mask_type)
        raise ValueError(errmsg)



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Definition of a masked waveform.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet
"""
import warnings

from fractions import Fraction
import numpy as np
import resampy
import simpleaudio as sa

from scipy.io import wavfile
from scipy.signal import hilbert
import matplotlib.pyplot as plt

from .mad_array import MadArray


VALID_IO_FS = {1, 8000, 16000, 32000, 48000, 11025, 22050, 44100, 88200}


def _check_compatibility_fs(w1, w2):
    """Raise an exception if the sampling frequency of the two Waveforms
    are different."""
    if w1.fs != w2.fs:
        errmsg = 'Waveforms do not have the same fs: {} and {}.'
        raise ValueError(errmsg.format(w1.fs, w2.fs))


class Waveform(MadArray):
    """Subclass of MadArray to handle mono and stereo audio signals.

    :class:`Waveform` inherits from :class:`~madarrays.mad_array.MadArray` and
    adds an attribute :attr:`fs` to store the sampling frequency, as well as
    methods to facilitate the manipulation of audio files.

    .. _type_entry_waveform:

    **Type of entries**: audio data entries can have different types, that are
    associated with specific constraints on the values:

    * *float*: the values are float between -1 and 1;
    * *int*: the values are integers between a range that depends on the
      precision.
    * *complex*: the real and imaginary parts are float between -1 and 1.

    High-precision types (float128, int64, complex256) may lead to problems and
    should be used with caution.

    The casting of a waveform in a different dtype depends on the current dtype
    and the desired dtype (complex casting is similar to float casting):

    * *Integer-to-real* casting is performed by applying on each entry
      :math:`x` the function :math:`f(x)=\\frac{x - z}{2^{n-1}}`, where the
      source integral type is coded with :math:`n` bits, and :math:`z` is the
      integer associated with zero, i.e., :math:`z=0` for a signed type (`int`)
      and :math:`z=2^{n-1}` for an unsigned type (`uint`).
    * *Real-to-integer* casting is performed by applying on each entry
      :math:`x` the function :math:`f(x)=\\lfloor\\left(x + 1\\right) 2^{n-1} +
      m\\rfloor`, where the target integral type is coded with :math:`n` bits,
      and :math:`m` is the minimum integer value, i.e., :math:`m=-2^{n-1}` for
      a signed type (`int`) and :math:`z=0` for an unsigned type (`uint`);
    * *Real-to-real* casting is obtained by a basic rounding operation;
    * *Integer-to-integer* casting is obtained by chaining an
      integer-to-float64 casting and a float64-to-integer casting.

    Clipping is performed for unexpected values:

    * When casting to `float`, values outside :math:`[-1, 1]` are clipped;
    * When casting to `int`, values outside the minimum and maximum values
      allowed by the integral type are clipped:

        * :math:`\\left[-2^{n-1}, 2^{n-1}-1\\right]` for :math:`n`-bits signed
          integers;
        * :math:`\\left[0, 2^{n}-1\\right]` for :math:`n`-bits unsigned
          integers.

    These constraints are only applied when calling explicitely the method
    :meth:`astype`.

    ..  _masking:

    **Masking**: Waveform allows for complex entries, but only real-like
    masking is permitted, i.e. it is not possible to mask only the phase or the
    amplitude. In particular, this implies that the attribute
    :attr:`_is_complex` inherited from :class:`~madarrays.mad_array.MadArray`
    is always equal to False.

    Parameters
    ----------
    data : nd-array [N] or [N, 2]
        Audio samples, as a N-length vector for a mono signal or a
        [N, 2]-shape array for a stereo signal
    fs : int or float, optional
        Sampling frequency of the original signal, in Hz. If float, truncated.
        If None and ``data`` is a Waveform, use ``data.fs``, otherwise it
        is set to 1.
    mask : nd-array, optional
        Boolean mask with True values for missing samples. Its shape must be
        the same as ``data``.
    indexing :
        See :class:`~madarrays.mad_array.MadArray`.
    """

    def __new__(cls, data, fs=None, mask=None, masked_indexing=None):
        if fs is None:
            if isinstance(data, Waveform):
                fs = data.fs
            else:
                fs = 1

        # create the MadArray
        data = data.squeeze()
        if isinstance(data, MadArray) and data._complex_masking \
                and mask is None :
            warnmsg = ('Construction of a Waveform from a MadArray with '
                       'complex masking: the original masking data are '
                       'discarded and a boolean mask set to False is used.')
            warnings.warn(warnmsg)
            # Create mask to force boolean masking
            mask = np.zeros_like(data, dtype=np.bool)

        obj = MadArray.__new__(cls=cls, data=data, mask=mask,
                               masked_indexing=masked_indexing)

        if obj.ndim > 2 or obj.ndim == 2 and obj.shape[1] > 2:
            raise ValueError('`data` should be either mono or stereo.')

        # add the new attribute to the created instance
        obj.fs = fs

        return obj

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):

        if len(inputs) == 2:
            if (isinstance(inputs[0], Waveform) and
                    isinstance(inputs[1], Waveform)):
                _check_compatibility_fs(*inputs)

        output = super().__array_ufunc__(ufunc, method, *inputs, **kwargs)

        if isinstance(output, MadArray):
            output = output.view(Waveform)
            output.fs = inputs[0].fs if isinstance(inputs[0], Waveform) \
                else inputs[1].fs

        return output

    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        self._fs = getattr(obj, '_fs', 1)
        self._play_object = getattr(obj, '_play_object', None)

    def __reduce__(self):
        pickled_state = super().__reduce__()
        new_state = pickled_state[2] + (self._fs, )
        return pickled_state[0], pickled_state[1], new_state

    def __setstate__(self, state):
        self._play_object = None
        self._fs = state[-1]
        super().__setstate__(state[0:-1])

    @property
    def fs(self):
        """Frequency sampling of the audio signal (int or float).

        The signal is not resampled when the sampling frequency is modified.

        Raises
        ------
        ValueError
            If `fs` is not positive.
        """
        return self._fs

    @fs.setter
    def fs(self, fs):
        if fs <= 0:
            errmsg = 'fs is not strictly positive (given: {})'
            raise ValueError(errmsg.format(fs))
        self._fs = fs

    @property
    def rms(self):
        """Root mean square error of the masked signal (float).

        Raises
        ------
        NotImplementedError
            If data is with an integer type.
        """
        if np.issubdtype(self.dtype, np.integer):
            errmsg = 'RMS is not available for integer data types'
            raise NotImplementedError(errmsg)

        return np.sqrt(np.mean(np.abs(self.to_np_array(0))**2))

    @property
    def length(self):
        """Length of the signal, in samples."""
        return self.shape[0]

    @property
    def duration(self):
        """Duration of the signal, in seconds."""
        return self.length / self.fs

    @property
    def n_channels(self):
        """Number of channels (1 for mono, 2 for stereo)."""
        return 1 if self.ndim == 1 else self.shape[1]

    @property
    def time_axis(self):
        """Time axis."""
        return np.arange(self.length) / self.fs

    def is_stereo(self):
        """Indicates whether the signal is stereo or not."""
        return self.n_channels == 2

    def set_rms(self, rms):
        """Set root mean square error of the signal.

        Parameters
        ----------
        rms : float
            Root mean square of the signal.

        Raises
        ------
        ValueError
            If `rms` is negative.
        NotImplementedError
            If data is with an integer type.
        """
        if rms < 0:
            errmsg = '`rms` should be a positive float (given: {})'
            raise ValueError(errmsg.format(rms))

        self[:] /= self.rms
        self[:] *= rms

    def resample(self, fs):
        """Resample the audio signal, in place.

        Can be only performed on a waveform without missing data.

        Note that if the current or the new sampling frequencies are not
        integers, the new sampling frequency may be different from the
        desired value since the resampling method only allows input and
        output frequencies of type ``int``. In this case, a warning is
        raised.

        Parameters
        ----------
        fs : int or float
            New sampling frequency.

        Raises
        ------
        ValueError
            If `fs` is not a positive number.
            If `self` has missing samples.
        UserWarning

        """
        assert np.issubdtype(type(fs), np.integer) \
               or np.issubdtype(type(fs), np.floating)

        if fs <= 0:
            errmsg = '`fs` should be a positive number (given: {})'
            raise ValueError(errmsg.format(fs))

        if np.issubdtype(type(fs), np.floating) \
                or np.issubdtype(type(self.fs), np.floating):
            # Find a good rational number to approximate the ratio between
            # sampling frequencies
            fs_ratio = Fraction(fs/self.fs)
            fs_ratio = fs_ratio.limit_denominator(10000)
            # Sampling frequencies used for the resampling (need to be int)
            resample_new_fs = fs_ratio.numerator
            resample_old_fs = fs_ratio.denominator
            # Adjust new sampling frequency
            new_fs = fs_ratio * self.fs
            if new_fs != fs:
                warnings.warn('New sampling frequency adjusted to {} instead '
                              'of {} in order to use ``int`` values in '
                              '``resample``.'.format(fs_ratio * self.fs, fs))
            fs = new_fs
        else:
            resample_new_fs = fs
            resample_old_fs = self.fs

        if self.is_masked():
            errmsg = 'Waveform has missing entries.'
            raise ValueError(errmsg)

        if resample_new_fs != resample_old_fs:
            x = self.to_np_array()

            y = resampy.resample(x, resample_old_fs, resample_new_fs, axis=0)

            self.resize(y.shape, refcheck=False)
            self[:] = y
            self._fs = fs
            self._mask = np.zeros_like(self, dtype=np.bool)

    def plot(self, x_axis_label='Time (s)', y_axis_label='',
             cpx_mode='real', fill_value=0, **kwargs):
        """Plot the signal.

        Parameters
        ----------
        x_axis_label : str
            Label of the x axis.
        y_axis_label : str
            Label of the y axis.
        cpx_mode : {'real', 'imag', 'both'}
            In case of a complex-valued signal, plot the real part only
            ('real'), the imaginary part only ('imag'), or both curves
            ('both'). This option has no effect if the signal is real-valued.
        fill_value : scalar or None
            Value used to fill missing data for display. If None, the existing
            value is used (e.g., for clipping).

        Other Parameters
        ----------------
        kwargs
            See :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        List of Line
            See :func:`matplotlib.pyplot.plot`.
        """
        x = self.to_np_array(fill_value)

        if np.issubdtype(x.dtype, np.complexfloating):
            if cpx_mode == 'real':
                lines = plt.plot(self.time_axis, np.real(x), **kwargs)
            elif cpx_mode == 'imag':
                lines = plt.plot(self.time_axis, np.imag(x), **kwargs)
            elif cpx_mode == 'both':
                lines_real = plt.plot(self.time_axis, np.real(x), **kwargs)
                lines_complex = plt.plot(
                    self.time_axis, np.imag(x), **kwargs)
                lines = lines_real + lines_complex
            else:
                errmsg = 'Unknown complex mode (given: {}).'
                raise ValueError(errmsg.format(cpx_mode))
        else:
            lines = plt.plot(self.time_axis, x, **kwargs)

        # add label_dict
        if x_axis_label is not None:
            plt.xlabel(x_axis_label)
        if y_axis_label is not None:
            plt.ylabel(y_axis_label)

        return lines

    def plot_mask(self,
                  x_axis_label='Time (s)',
                  y_axis_label='',
                  **kwargs):
        """Plot the mask.

        Parameters
        ----------
        x_axis_label : str
            Label of the time axis (horizontal axis).
        y_axis_label : str
            Label of the frequency axis (vertical axis).
        """
        plt.step(
            self.time_axis, self._mask, linewidth=2, where='pre', **kwargs)
        plt.ylim(-0.25, 1.25)
        plt.yticks([0, 1])

        plt.xlabel(x_axis_label)
        plt.ylabel(y_axis_label)
        plt.grid()

    def to_wavfile(self, filename, dtype=None):
        """Save the wavefile as an audio wav file.

        Parameters
        ----------
        filename : str or :class:`pathlib.Path`
            Name of the file including paths and extension
        dtype : data-type or None, optional
            Type of the entries of the wav file. If None, use current dtype of
            data.

        Raises
        ------
        UserWarning
            If the signal is complex.
        ValueError
            If the sampling frequency is not an integer from the set of
            supported frequencies ``madarrays.waveform.VALID_IO_FS``).
        NotImplementedError
            If dtype is not supported by the current implementation.

        See also
        --------
        scipy.io.wavfile.write
        """
        if self.fs not in VALID_IO_FS:
            errmsg = '`fs` is not a valid sampling frequency (given: {}).'
            raise ValueError(errmsg.format(self.fs))

        if dtype is None:
            dtype = self.dtype

        y = self.to_np_array(0)

        if np.iscomplexobj(y):
            warnmsg = 'Complex values have been stored by keeping only the '\
                'real part, the imaginary part has been discarded.'
            warnings.warn(warnmsg)

            y = np.real(y)

            if np.issubdtype(dtype, np.complexfloating):
                dtype = np.float32

        if dtype == np.float64:
            dtype = np.float32
        elif dtype == np.int64:
            raise NotImplementedError

        # catch clipping error
        y = y.astype(dtype=dtype)

        if not np.all(np.isfinite(y)):
            errmsg = 'Data should be finite.'
            raise ValueError(errmsg)

        wavfile.write(str(filename), data=y, rate=self.fs)

    @staticmethod
    def from_wavfile(filename, dtype=np.float64, conversion_to_mono=None):
        """Load an audio file and return an Waveform object.

        Parameters
        ----------
        filename : str or :class:`pathlib.Path`
            Name of the audio file
        dtype : data-type, optional
            Output data type. If None, the data type from the wavfile is kept.
            See :ref:`Type of Entry<type_entry_waveform>`.
        conversion_to_mono : {'left', 'right', 'mean'} or None, optional
            If None (default), no conversion is performed. If str, select the
            appropriate channel (left or right), or the mean between the left
            and right channels if stereo.

        Returns
        -------
        :class:`Waveform`
        """
        fs, x = wavfile.read(str(filename))

        if dtype is None:
            dtype = x.dtype

        if x.ndim > 1:
            if conversion_to_mono is not None:

                if conversion_to_mono == 'left':
                    x = x[:, 0]

                elif conversion_to_mono == 'right':
                    x = x[:, 1]

                elif conversion_to_mono == 'mean':
                    x = np.mean(x, axis=1)

                else:
                    errmsg = '`conversion_to_mono` in invalid (given: {}).'
                    raise ValueError(errmsg.format(conversion_to_mono))

        return Waveform(x, fs=fs).astype(dtype=dtype)

    def fade(self, mode='both', fade_duration=None, fade_length=None):
        """Apply fade-in and/or fade-out.

        Half a hanning window with the specified length or duration is used
        (exactly one parameter `fade_duration` or `fade_length` must be set).

        Parameters
        ----------
        mode : {'both', 'in', 'out'}, optional
            Signal to be processed.
        fade_duration : float, optional
            Duration, in seconds, of the fade effect.
        fade_length : int, optional
            Length, in samples, of the fade effect.

        Raises
        ------
        ValueError
            If neither `fade_duration` nor `fade_length` is given.
            If both `fade_duration` and `fade_length` are given.
            If `fade_duration` or `fade_length` are greater than the
            duration or length of the waveform.
            If `fade_duration` or `fade_length` are negative.
            If `mode` is invalid.
        """
        if fade_duration is None and fade_length is None:
            errmsg = 'Either `fade_duration` or `fade_length` must be given.'
            raise ValueError(errmsg)

        if fade_duration is not None and fade_length is not None:
            errmsg = '`fade_duration` and `fade_length` cannot be given '\
                'simultaneously.'
            raise ValueError(errmsg)

        if mode not in ['both', 'in', 'out']:
            errmsg = 'Mode is unknown: {}.'
            raise ValueError(errmsg.format(mode))

        if fade_length is None:
            fade_length = int(np.round(fade_duration * self.fs))

        if fade_length == 0:
            return

        if fade_length < 0:
            errmsg = 'Fade length cannot be negative.'
            raise ValueError(errmsg)

        if fade_length > self.shape[0]:
            errmsg = 'Fade length ({}) is greater than data length ({})'
            raise ValueError(errmsg.format(fade_length, self.shape[0]))

        w = np.sin(np.pi / 2 / fade_length * np.arange(fade_length))**2

        if mode in ['both', 'in']:
            self[:len(w)] *= w

        if mode in ['both', 'out']:
            self[-len(w):] *= w[-1::-1]

    def show_player(self):
        """
        Show the player when using jupyter notebook.

        Raises
        ------
        ValueError
            If sampling frequency is 1.
        """
        try:
            from IPython.display import Audio
            if self.fs == 1:
                errmsg = 'Invalid sampling frequency: {}Hz'
                raise ValueError(errmsg.format(self.fs))

            if np.iscomplexobj(self):
                warnmsg = 'Only the real part is played.'
                warnings.warn(warnmsg)

            x = np.real(self).to_np_array(0)
            return Audio(data=x, rate=self.fs)
        except ImportError:
            print('This method should only be called in notebooks.')

    def play(self, scale=False):
        """Play the audio file.

        Parameters
        ----------
        scale : bool
            If True, the data is normalized by maximum value.

        Returns
        -------
        `simpleaudio.PlayObject`
            Object useful to handle the played audio stream.

        Raises
        ------
        ValueError
            If sampling frequency is 1.
        """

        if self.fs == 1:
            errmsg = 'Invalid sampling frequency: {}Hz'
            raise ValueError(errmsg.format(self.fs))

        if np.iscomplexobj(self):
            warnmsg = 'Only the real part is played.'
            warnings.warn(warnmsg)

        x = np.real(self)

        if scale:
            x = x.astype(np.float64)
            maxima = np.max(np.abs(self))
            if maxima > 0:
                x[:] = x / maxima * (1 - np.finfo(np.float64).eps)

        x = x.astype(dtype=np.int16).to_np_array(0)
        self._play_object = sa.play_buffer(x, num_channels=self.n_channels,
                                           bytes_per_sample=2,
                                           sample_rate=self.fs)

    def stop(self):
        """Stop playing (does nothing if the sound is not played) """
        if self.is_playing():
            self._play_object.stop()

    def is_playing(self):
        """Indicates whether the sound is currently being played.

        Returns
        -------
        bool
            True if the sound is currently being played, False otherwise.
        """
        return (hasattr(self._play_object, 'is_playing') and
                self._play_object.is_playing())

    def get_analytic_signal(self, N=None, axis=0):
        """Return the analytic waveform.

        Parameters
        ----------
        N : int, optional
            Number of Fourier components. If None, set to x.shape[axis].
        axis : int, optional
            Axis along which to do the transformation.

        Returns
        -------
        Waveform

        Raises
        ------
        TypeError
            If Waveform has missing samples.

        See Also
        --------
        scipy.signal.hilbert
        """
        if np.any(self.get_unknown_mask()):
            errmsg = 'Waveform has missing samples.'
            raise TypeError(errmsg)

        y = hilbert(self, N=N, axis=axis)
        return Waveform(data=y, fs=self.fs)

    def clip(self, min_value=None, max_value=None):
        """Clip audio data according to given bound values.

        See :ref:`Type of Entry<type_entry_waveform>` for details.

        Parameters
        ----------
        max_value : float, optional
            Clipping is applied above this value.
        min_value : float, optional
            Clipping is applied below this value.

        Raises
        ------
        UserWarning
            If some coefficients are actually clipped.

        Notes
        -----
        Masked values can be also clipped.
        """
        if min_value is None:
            min_value = np.min(self)

        if max_value is None:
            max_value = np.max(self)

        if np.issubdtype(self.dtype, np.complexfloating):
            w_real = Waveform(np.real(self))
            w_imag = Waveform(np.imag(self))

            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                w_real.clip(min_value=min_value, max_value=max_value)

            warnmsg = 'Real part of the complex entries: {}'
            for w_i in w:
                warnings.warn(warnmsg.format(str(w_i.message)))

            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                w_imag.clip(min_value=min_value, max_value=max_value)

            warnmsg = 'Imaginary part of the complex entries: {}'
            for w_i in w:
                warnings.warn(warnmsg.format(str(w_i.message)))

            self[:] = np.array(w_real) + 1j * np.array(w_imag)

        else:
            ind_min = self < min_value
            if np.any(ind_min):
                self[ind_min] = min_value
                warnmsg = '{} values lower than {} have been clipped.'
                warnings.warn(warnmsg.format(self.dtype, min_value))

            ind_max = self > max_value
            if np.any(ind_max):
                self[ind_max] = max_value
                warnmsg = '{} values greater than {} have been clipped.'
                warnings.warn(warnmsg.format(self.dtype, max_value))

    def astype(self, dtype):
        """Cast the waveform into a given type (float, int, or complex).

        Parameters
        ----------
        dtype : data-type, optional
            Output data type. See :ref:`Type of Entry<type_entry_waveform>`.

        Returns
        -------
        nd-array
            Waveform with the desired dtype.

        Raises
        ------
        TypeError
            If argument `dtype` is an unsupported data type.
        UserWarning
            If some coefficients are actually clipped.
        """
        y = self.copy()

        if (np.issubdtype(self.dtype, np.floating) or
                np.issubdtype(self.dtype, np.complexfloating)):

            if (np.issubdtype(dtype, np.floating) or
                    np.issubdtype(dtype, np.complexfloating)):
                y.clip(min_value=-1, max_value=(1 - np.finfo(dtype).eps))
                y = np.asarray(y, dtype=dtype)

            elif np.issubdtype(dtype, np.integer):
                target_type_info = np.iinfo(dtype)
                int_range = target_type_info.max - target_type_info.min + 1
                y = np.floor((np.real(y) + 1) / 2 * int_range +
                             target_type_info.min)
                y.clip(min_value=np.iinfo(dtype).min,
                       max_value=np.iinfo(dtype).max)
                y = np.asarray(y, dtype=dtype)

            else:
                errmsg = 'Unsupported target type {}.'
                raise TypeError(errmsg.format(dtype))

        else:
            src_iinfo = np.iinfo(y.dtype)

            if np.issubdtype(dtype, np.floating):
                int_range = src_iinfo.max - src_iinfo.min + 1
                zero = src_iinfo.min + int_range // 2
                y = ((super(type(y), y).astype(dtype) - zero)
                     / (int_range // 2))
                y.clip(min_value=-1, max_value=1-np.finfo(dtype).eps)
                y = np.asarray(y, dtype=dtype)

            elif np.issubdtype(dtype, np.integer):
                y = y.astype(dtype=np.float64)
                y = y.astype(dtype=dtype)
                y = np.asarray(y)

            elif np.issubdtype(dtype, np.complexfloating):
                y = y.astype(dtype=np.float64)
                y = np.asarray(y, dtype=dtype)

            else:
                errmsg = 'Unsupported target type {}.'
                raise TypeError(errmsg.format(dtype))

        return Waveform(y, mask=self.get_unknown_mask(), fs=self.fs)

    def copy(self):
        return Waveform(self)

    def __eq__(self, other):
        if isinstance(other, Waveform):
            _check_compatibility_fs(self, other)

        return super().__eq__(other)

    def is_equal(self, other):
        return self.fs == other.fs and super().is_equal(other)

    def __str__(self):

        string = super().__str__()
        string = string.replace('MadArray', 'Waveform, fs={}Hz, length={}')

        return string.format(self.fs, self.length)

    def __repr__(self):
        string = super().__repr__()
        string = string.replace('MadArray', 'Waveform')
        return string



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Utils functions for testing.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet
"""
import numpy as np


def generate_mask_50(shape):

    n_elements = np.prod(shape)
    m = np.zeros(shape, dtype=np.bool)
    ind = np.random.choice(np.arange(n_elements), n_elements // 2,
                           replace=False)

    m.flat[ind] = True
    return m


def assert_array_less_or_equal(x, y, **kwargs):
    eps = np.finfo(x.dtype).eps
    np.testing.assert_array_less(x - eps, y, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys
import os

from datetime import date

import madarrays

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../madarrays/'))

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.doctest',
              'sphinx.ext.intersphinx',
              'sphinx.ext.todo',
              'sphinx.ext.coverage',
              'sphinx.ext.mathjax',
              'sphinx.ext.viewcode',
              'numpydoc',
              'nbsphinx',
              'IPython.sphinxext.ipython_console_highlighting',
              ]


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'madarrays'
author = 'V. Emiya, R. Hamon, F. Jaillet'
copyright = '2017-{}, {}'.format(date.today().year, author)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = madarrays.__version__
# The full version, including alpha/beta/rc tags.
release = madarrays.__version__.replace('_', '')

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '**/test_*.rst', '**.ipynb_checkpoints']
# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'bizstyle'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# Custom sidebar templates, maps document names to template names.
html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr'
html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
html_search_scorer = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'madarraydoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    'papersize': 'a4paper',

    # The font size ('10pt', '11pt' or '12pt').
    'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    'preamble': '',

    # Latex figure (float) alignment
    'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'madarray.tex', 'madarray Documentation',
     'V. Emiya, R. Hamon., F. Jaillet', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
latex_use_parts = False

# If true, show page references after internal links.
latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
latex_appendices = []

# If false, no module index is generated.
latex_domain_indices = True

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'madarray', 'madarray Documentation', [author], 1)]

# If true, show URL addresses after external links.
man_show_urls = False

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'madarray', 'madarray Documentation', author, 'madarray',
     'One line description of project.', 'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
texinfo_appendices = []

# If false, no module index is generated.
texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
texinfo_no_detailmenu = False

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('http://matplotlib.org/', None),
}

# Allow errors in notebook
nbsphinx_allow_errors = True

# Do not show class members
numpydoc_show_class_members = False

# Include todos
todo_include_todos = True

# Keep the order of members as in the source code
autodoc_member_order = 'bysource'



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes: http://www.lis-lab.fr/
# * Université d'Aix-Marseille : http://www.univ-amu.fr/
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# madarrays is a Python module for audio data structures with missing entries
#
# Version
# -------
#
# * madarrays version=1.0
#
# Licence
# -------
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Data structures adapted to handle audio signals with missing data

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet
"""

from .mad_array import MadArray
from .waveform import Waveform

__all__ = ['MadArray', 'Waveform']
__version__ = "1.1.0"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
import os
from setuptools import setup, find_packages
import sys

NAME = 'madarrays'
DESCRIPTION = 'Python package for audio data structures with missing entries'
LICENSE = 'GNU General Public License v3 (GPLv3)'
URL = 'https://gitlab.lis-lab.fr/skmad-suite/{}'.format(NAME)
AUTHOR = 'Ronan Hamon, Valentin Emiya, and Florent Jaillet'
AUTHOR_EMAIL = ('ronan.hamon@lis-lab.fr, valentin.emiya@lis-lab.fr, '
                'florent.jaillet@lis-lab.fr')
INSTALL_REQUIRES = ['matplotlib',
                    'numpy',
                    'scipy',
                    'simpleaudio',
                    'resampy']
CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Science/Research',
    'Topic :: Multimedia :: Sound/Audio :: Analysis',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Natural Language :: English',
    'Operating System :: MacOS :: MacOS X ',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6']
PYTHON_REQUIRES = '>=3.5'
EXTRAS_REQUIRE = {
    'dev': ['coverage', 'pytest', 'pytest-cov', 'pytest-randomly', 'ipython'],
    'doc': ['sphinx', 'nbsphinx', 'numpydoc']}
PROJECT_URLS = {'Bug Reports': URL + '/issues',
                'Source': URL}
KEYWORDS = 'audio, data structures, missing data'

###############################################################################
if sys.argv[-1] == 'setup.py':
    print("To install, run 'python setup.py install'\n")

if sys.version_info[:2] < (3, 5):
    errmsg = '{} requires Python 3.5 or later ({[0]:d}.{[1]:d} detected).'
    print(errmsg.format(NAME, sys.version_info[:2]))
    sys.exit(-1)


def get_version():
    v_text = open('VERSION').read().strip()
    v_text_formted = '{"' + v_text.replace('\n', '","').replace(':', '":"')
    v_text_formted += '"}'
    v_dict = eval(v_text_formted)
    return v_dict[NAME]


def set_version(path, VERSION):
    filename = os.path.join(path, '__init__.py')
    buf = ""
    for line in open(filename, "rb"):
        if not line.decode("utf8").startswith("__version__ ="):
            buf += line.decode("utf8")
    f = open(filename, "wb")
    f.write(buf.encode("utf8"))
    f.write(('__version__ = "%s"\n' % VERSION).encode("utf8"))


def setup_package():
    """Setup function"""
    # set version
    VERSION = get_version()

    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()

    mod_dir = NAME
    set_version(mod_dir, get_version())
    setup(name=NAME,
          version=VERSION,
          description=DESCRIPTION,
          long_description=long_description,
          url=URL,
          author=AUTHOR,
          author_email=AUTHOR_EMAIL,
          license=LICENSE,
          classifiers=CLASSIFIERS,
          keywords=KEYWORDS,
          packages=find_packages(exclude=['doc', 'dev']),
          install_requires=INSTALL_REQUIRES,
          python_requires=PYTHON_REQUIRES,
          extras_require=EXTRAS_REQUIRE,
          project_urls=PROJECT_URLS)


if __name__ == "__main__":
    setup_package()



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# Python package for audio data structures with missing entries
#
# Licence
# -------
# This file is part of madarrays.
#
# madarrays is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
